/*
  FSaccelGyro.cpp - FreeSensors library for Arduino
  Copyright (C) 2018 Mario A. Guerra. All rights reserved.
  License: GNU GPL v3 http://www.gnu.org/licenses/gpl.html

  Accelerometer-Gyroscope for FreeSensors:
  based on the MPU6050_tockn library: easy communicating with the MPU6050
  https://github.com/Tockn/MPU6050_tockn  version=1.4.0
*/

#include "FSaccelGyro.h"

#define MPU6050_Address0     0x68
#define MPU6050_Address1     0x69
#define MPU6050_SMPLRT_DIV   0x19
#define MPU6050_CONFIG       0x1A
#define MPU6050_GYRO_CONFIG  0x1B
#define MPU6050_ACCEL_CONFIG 0x1C
#define MPU6050_RAW_DATA     0x3B
#define MPU6050_PWR_MGMT_1   0x6B
#define CALIBRATION_SAMPLES  1500
#define ACCEL_RANGE_2G       B00000000
#define ACCEL_RANGE_4G       B00001000
#define ACCEL_RANGE_8G       B00010000
#define ACCEL_RANGE_16G      B00011000
#define GYRO_RANGE_250       B00000000
#define GYRO_RANGE_500       B00001000
#define GYRO_RANGE_1000      B00010000
#define GYRO_RANGE_2000      B00011000

FSaccelGyro::FSaccelGyro() {
    mtype = fmtI2C;
    defaultAddr = MPU6050_Address0;
    accCoef = 0.04;
    gyroCoef = 0.96;
    accelXoffset = 0;
    accelYoffset = 0;
    accelZoffset = 0;
    gyroXoffset = 0;
    gyroYoffset = 0;
    gyroZoffset = 0;
}

void FSaccelGyro::begin() {
    attach(defaultAddr);
    initialize();
}

void FSaccelGyro::initialize() {
    writeConfig(MPU6050_SMPLRT_DIV, 0x00);
    writeConfig(MPU6050_CONFIG, 0x00);
    writeConfig(MPU6050_ACCEL_CONFIG, ACCEL_RANGE_2G);
    writeConfig(MPU6050_GYRO_CONFIG, GYRO_RANGE_250);
    writeConfig(MPU6050_PWR_MGMT_1, 0x01);  // CLKSEL = PLL with X Gyro
    rangeAccel = 2;    // Range up to 2G
    rangeGyro = 250;   // Range up to 250 º/s
    setZeroPosition();
    prevTime = millis();
}

void FSaccelGyro::readRawValues() {
    Wire.beginTransmission(deviceAddr);
    Wire.write(MPU6050_RAW_DATA);
    Wire.endTransmission(false);
    Wire.requestFrom(deviceAddr, (byte)14);  // Request a total of 14 registers
    rawAccX =  Wire.read() << 8 | Wire.read();  // 0x3B (ACCEL_XOUT_H) & 0x3C (ACCEL_XOUT_L)
    rawAccY =  Wire.read() << 8 | Wire.read();  // 0x3D (ACCEL_YOUT_H) & 0x3E (ACCEL_YOUT_L)
    rawAccZ =  Wire.read() << 8 | Wire.read();  // 0x3F (ACCEL_ZOUT_H) & 0x40 (ACCEL_ZOUT_L)
    rawTemp =  Wire.read() << 8 | Wire.read();  // 0x41 (TEMP_OUT_H)   & 0x42 (TEMP_OUT_L)
    rawGyroX = Wire.read() << 8 | Wire.read();  // 0x43 (GYRO_XOUT_H)  & 0x44 (GYRO_XOUT_L)
    rawGyroY = Wire.read() << 8 | Wire.read();  // 0x45 (GYRO_YOUT_H)  & 0x46 (GYRO_YOUT_L)
    rawGyroZ = Wire.read() << 8 | Wire.read();  // 0x47 (GYRO_ZOUT_H)  & 0x48 (GYRO_ZOUT_L)
    Wire.endTransmission();
    DBG_PRINT("raw AccX = "); DBG_PRINTLN(rawAccX);
    DBG_PRINT("raw AccY = "); DBG_PRINTLN(rawAccY);
    DBG_PRINT("raw AccZ = "); DBG_PRINTLN(rawAccZ);
    DBG_PRINT("raw GyroX = "); DBG_PRINTLN(rawGyroX);
    DBG_PRINT("raw GyroY = "); DBG_PRINTLN(rawGyroY);
    DBG_PRINT("raw GyroZ = "); DBG_PRINTLN(rawGyroZ);
}

void FSaccelGyro::updateValues() {
    const float convertDG = 180 / PI;
    const float GRAV = 9.80665;
    float interval;
    unsigned long currTime;

    readRawValues();
    currTime = millis();
    interval = (currTime - prevTime) / 1000.0;  // interval is in seconds
    prevTime = currTime;

    // Acceleration with values in G units, speed in [m/s], and angles in degrees
    accX = (float)rawAccX * rangeAccel / 32768 - accelXoffset;
    accY = (float)rawAccY * rangeAccel / 32768 - accelYoffset;
    accZ = (float)rawAccZ * rangeAccel / 32768 - accelZoffset;
    speedX = speedX + accX * GRAV * interval;
    speedY = speedY + accY * GRAV * interval;
    speedZ = speedZ + accZ * GRAV * interval;
    angleAccX = atan2(accY, accZ + abs(accX)) * convertDG;
    angleAccY = atan2(accX, accZ + abs(accY)) * -convertDG;

    // Angular velocity with values in degree/second units, and angles in degrees
    gyroX = (float)rawGyroX * rangeGyro / 32768 - gyroXoffset;
    gyroY = (float)rawGyroY * rangeGyro / 32768 - gyroYoffset;
    gyroZ = (float)rawGyroZ * rangeGyro / 32768 - gyroZoffset;
    angleGyroX = angleGyroX + gyroX * interval;
    angleGyroY = angleGyroY + gyroY * interval;
    angleGyroZ = angleGyroZ + gyroZ * interval;

    // Angles in degrees using complementary filter for X and Y
    angleX = (gyroCoef * (angleX + gyroX * interval)) + (accCoef * angleAccX);
    angleY = (gyroCoef * (angleY + gyroY * interval)) + (accCoef * angleAccY);
    angleZ = angleGyroZ;
}

void FSaccelGyro::setCoefficients(float accC, float gyrC) {
    accCoef = accC;
    gyroCoef = gyrC;
}

void FSaccelGyro::setRanges(short accR, short gyrR) {
    byte aRange, gRange;

    switch (accR) {
        case 2:  aRange = ACCEL_RANGE_2G; break;
        case 4:  aRange = ACCEL_RANGE_4G; break;
        case 8:  aRange = ACCEL_RANGE_8G; break;
        case 16: aRange = ACCEL_RANGE_16G; break;
        default: aRange = ACCEL_RANGE_2G;
                 accR = 2;
    }
    rangeAccel = accR;
    writeConfig(MPU6050_ACCEL_CONFIG, aRange);
    switch (gyrR) {
        case 250:  gRange = GYRO_RANGE_250; break;
        case 500:  gRange = GYRO_RANGE_500; break;
        case 1000: gRange = GYRO_RANGE_1000; break;
        case 2000: gRange = GYRO_RANGE_2000; break;
        default:   gRange = GYRO_RANGE_250;
                   gyrR = 250;
    }
    rangeGyro = gyrR;
    writeConfig(MPU6050_GYRO_CONFIG, gRange);
}

void FSaccelGyro::setAccelOffsets(float x, float y, float z) {
    accelXoffset = x;
    accelYoffset = y;
    accelZoffset = z;
}

void FSaccelGyro::setGyroOffsets(float x, float y, float z) {
    gyroXoffset = x;
    gyroYoffset = y;
    gyroZoffset = z;
}

void FSaccelGyro::setZeroPosition() {
    speedX = 0;
    speedY = 0;
    speedZ = 0;
    angleGyroX = 0;
    angleGyroY = 0;
    angleGyroZ = 0;
    angleX = 0;
    angleY = 0;
    angleZ = 0;
}

void FSaccelGyro::calibrate(bool console /* = false */) {
    float sumAX, sumAY, sumAZ;
    float sumGX, sumGY, sumGZ;

    delay(400);
    if (console) {
        Serial.println();
        Serial.println("========================================");
        Serial.println("Calculating offsets");
        Serial.print("DO NOT MOVE the MPU6050 ");
    }
    sumAX = 0;
    sumAY = 0;
    sumAZ = 0;
    sumGX = 0;
    sumGY = 0;
    sumGZ = 0;
    for (int i = 1; i <= CALIBRATION_SAMPLES; i++) {
        if (console && i % 200 == 0) Serial.print(".");
        readRawValues();
        sumAX = sumAX + (float)rawAccX * rangeAccel / 32768;
        sumAY = sumAY + (float)rawAccY * rangeAccel / 32768;
        sumAZ = sumAZ + (float)rawAccZ * rangeAccel / 32768;
        sumGX = sumGX + (float)rawGyroX * rangeGyro / 32768;
        sumGY = sumGY + (float)rawGyroY * rangeGyro / 32768;
        sumGZ = sumGZ + (float)rawGyroZ * rangeGyro / 32768;
        delay(2);   // needed so we don't get repeated measures
    }
    accelXoffset = sumAX / CALIBRATION_SAMPLES;
    accelYoffset = sumAY / CALIBRATION_SAMPLES;
    accelZoffset = sumAZ / CALIBRATION_SAMPLES - 1; // Normal is 1G
    gyroXoffset = sumGX / CALIBRATION_SAMPLES;
    gyroYoffset = sumGY / CALIBRATION_SAMPLES;
    gyroZoffset = sumGZ / CALIBRATION_SAMPLES;
    setZeroPosition();
    if (console) {
        Serial.println();
        Serial.println("Done !!");
        Serial.print("Offset Accel X : "); Serial.println(accelXoffset);
        Serial.print("Offset Accel Y : "); Serial.println(accelYoffset);
        Serial.print("Offset Accel Z : "); Serial.println(accelZoffset);
        Serial.print("Offset Gyro X : "); Serial.println(gyroXoffset);
        Serial.print("Offset Gyro Y : "); Serial.println(gyroYoffset);
        Serial.print("Offset Gyro Z : "); Serial.println(gyroZoffset);
        Serial.println("Program will start after 2 seconds");
        Serial.println("========================================");
        Serial.println();
        delay(2000);
    }
}
