/*
  FSmagCompass.cpp - FreeSensors library for Arduino
  Copyright (C) 2018 Mario A. Guerra. All rights reserved.
  License: GNU GPL v3 http://www.gnu.org/licenses/gpl.html

  Magnetometer-Compass for FreeSensors:
  based on the QMC5883L Triple Axis Magnetometer Arduino Library.
  Copyright (C) 2017 Andy Barnard based on an original HMC5883L by Love Electronics
*/

#include "FSmagCompass.h"

#define QMC5883L_Address  0x0D
#define ControlRegister1  0x09
#define ControlRegister2  0x0A
#define RstPeriodRegister 0x0B
#define DataRegisterBegin 0x00  // QMC5883 from 00H to 05H
#define TempRegisterBegin 0x07  // QMC5883 from 07H to 08H

#define OVERSAMP_RATE_512 B00   // Over sample Rate 512
#define OVERSAMP_RATE_256 B01   // Over sample Rate 256
#define OVERSAMP_RATE_128 B10   // Over sample Rate 128
#define OVERSAMP_RATE_064 B11   // Over sample Rate 64
#define MAGN_RANGE_2G     B00   // Full scale range +/- 2 Gauss
#define MAGN_RANGE_8G     B01   // Full scale range +/- 8 Gauss
#define OUT_DATA_RATE_010 B00   // Output data rate 10 Hz
#define OUT_DATA_RATE_050 B01   // Output data rate 50 Hz
#define OUT_DATA_RATE_100 B10   // Output data rate 100 Hz
#define OUT_DATA_RATE_200 B11   // Output data rate 200 Hz
#define CTRL_MODE_STANDBY B00   // Standby mode
#define CTRL_MODE_CONTIN  B01   // Continuous measurement mode
#define CR2_INT_DISABLE   B00000000 // Control register 2: disable soft reset and pointer rollover, interrupt pin not enabled
#define RESET_PERIOD_REG  B00000001 // datasheet recommends register be written with 1

FSmagCompass::FSmagCompass() {
    mtype = fmtI2C;
    defaultAddr = QMC5883L_Address;
    declination = 0;
    heading = 0;
}

void FSmagCompass::begin() {
    attach(defaultAddr);
    initialize();
}

void FSmagCompass::initialize() {
    byte CR1config;
    CR1config = (OVERSAMP_RATE_512 << 6) | (MAGN_RANGE_2G << 4)  | (OUT_DATA_RATE_010 << 2) | CTRL_MODE_CONTIN;
    writeConfig(ControlRegister1, CR1config);
    writeConfig(ControlRegister2, CR2_INT_DISABLE);
    writeConfig(RstPeriodRegister, RESET_PERIOD_REG);
    rangeGauss = 2;    // Full scale range = +/- 2 Gauss
}

void FSmagCompass::readRawValues() {
    int byteLSB;
    int byteMSB;

    Wire.beginTransmission(deviceAddr);
    Wire.write(DataRegisterBegin);
    Wire.endTransmission(false);
    Wire.requestFrom(deviceAddr, (byte)6);
    // Request a total of 6 registers, for each axis first is LSB, second is MSB
    byteLSB = Wire.read();
    byteMSB = Wire.read();
    rawAxisX =  byteMSB << 8 | byteLSB;
    byteLSB = Wire.read();
    byteMSB = Wire.read();
    rawAxisY =  byteMSB << 8 | byteLSB;
    byteLSB = Wire.read();
    byteMSB = Wire.read();
    rawAxisZ =  byteMSB << 8 | byteLSB;
    Wire.endTransmission();
    DBG_PRINT("raw X = "); DBG_PRINTLN(rawAxisX);
    DBG_PRINT("raw Y = "); DBG_PRINTLN(rawAxisY);
    DBG_PRINT("raw Z = "); DBG_PRINTLN(rawAxisZ);
}

void FSmagCompass::updateValues() {
    float radians, mag_north, mag_west;

    readRawValues();
    magnX = (float)rawAxisX * rangeGauss / 32768;
    magnY = (float)rawAxisY * rangeGauss / 32768;
    magnZ = (float)rawAxisZ * rangeGauss / 32768;
    mag_north = magnX;
    mag_west = magnY;
    radians = atan2(mag_west, mag_north);         // calculate heading from north and west magnetic axes
    radians = radians - declination;              // adjust by declination (a positive (E) must be substracted)
    if (radians < 0) radians = radians + 2*PI;    // correct for when angle is negative
    if (radians > 2*PI) radians = radians - 2*PI; // check for wrap due to declination addition
    heading = radians * 180 / PI;                 // convert radians to degrees for readability
}

void FSmagCompass::setRange(short range) {
    byte CR1config, mRange;

    switch (range) {
        case 2:  mRange = MAGN_RANGE_2G; break;
        case 8:  mRange = MAGN_RANGE_8G; break;
        default: mRange = MAGN_RANGE_2G;
                 range = 2;
    }
    rangeGauss = range;
    CR1config = (OVERSAMP_RATE_512 << 6) | (mRange << 4)  | (OUT_DATA_RATE_010 << 2) | CTRL_MODE_CONTIN;
    writeConfig(ControlRegister1, CR1config);
}

void FSmagCompass::setDeclination(float radians) {
    declination = radians;
}

// Convert declination to decimal degrees and then to radians.
// A positive declination (toward East) will be substracted from heading, and
// a negative declination (toward West) will be added to heading.
void FSmagCompass::setDeclination(int degs, int mins, char direc) {
    switch (direc) {
        case 'E': declination = (degs + (mins/60)) * PI / 180; break;
        case 'W': declination = - (degs + (mins/60)) * PI / 180; break;
        default : declination = 0;
    }
}
