/*
  FSmotor.cpp - FreeSensors library for Arduino.
  Copyright (C) 2019 Mario A. Guerra. All rights reserved.
  This file is part of FreeSensors, licensed under GNU GPL version 3.
*/

#include "FSmotor.h"

#define ENABLE_DCM1 5      // Arduino digital pin
#define ENABLE_DCM2 6      // Arduino digital pin
#define DCM1_PLUS 5        // bit in I2C register
#define DCM1_MINU 6        // bit in I2C register
#define DCM2_PLUS 7        // bit in I2C register
#define DCM2_MINU 8        // bit in I2C register

FSmotor::FSmotor() {
    stype = fstMOTOR;
}

void FSmotor::begin() {
    begin(PCF_010);
}

void FSmotor::begin(byte address) {
    port[0].jack = 1;    port[0].type = fptDIOF;
    port[1].jack = 2;    port[1].type = fptDIOF;
    port[2].jack = 3;    port[2].type = fptDIOF;
    port[3].jack = 4;    port[3].type = fptDIOF;
    port[4].jack = 1;    port[4].type = fptAO;     port[4].pin = 9;
    port[5].jack = 2;    port[5].type = fptAO;     port[5].pin = 10;
    port[6].jack = 1;    port[6].type = fptDC;
    port[7].jack = 2;    port[7].type = fptDC;
    port[8].jack = 1;    port[8].type = fptSER;
    port[9].jack = 1;    port[9].type = fptI2C;
    attach(address);
    registro = 0;
    writeValue();
    pinModeFS(ENABLE_DCM1, OUTPUT);  // PWM may control speed
    pinModeFS(ENABLE_DCM2, OUTPUT);  // PWM may control speed
    connect(m1,1);
    connect(m2,2);
}

// This connect is needed to differentiate between FSshield  and FSmodule connects
bool FSmotor::connect(FSmodule &xmod, int xjack) {
    return FSshield::connect(xmod, xjack);
}

void FSmotor::setPinMode(byte jack, byte mode) {
    switch (mode) {
        case OUTPUT: bitWrite(direccion,jack-1,0); break;
        case INPUT:
        case INPUT_PULLUP: bitWrite(direccion,jack-1,1); break;
    }
}

bool FSmotor::readDigitalValue(byte pin) {
    if ((pin<1) || (pin>8)) return 0;
    readValue();
    return bitRead(registro,pin-1);
}

void FSmotor::writeDigitalValue(byte pin, byte value) {
    if ((pin<1) || (pin>8)) return;
    // On input pins always write 1
    registro = registro | direccion;
    bitWrite(registro,pin-1,value);
    writeValue();
}

void FSmotor::writePwmValue(byte jack, byte value) {
    analogWriteFS(port[jack+3].pin, value);
}

void FSmotor::setPinOutput(byte jack) {
    pinModeFS(port[jack+3].pin, OUTPUT);
}

void FSmotor::doPulse(byte jack, byte duration) {
    digitalWriteFS(port[jack+3].pin, HIGH);
    delayMicroseconds(duration);
    digitalWriteFS(port[jack+3].pin, LOW);
}

void FSmotor::startServo(byte jack) {
    // Using the "Servo" library initializes an array of 12 servos structs,
    // the constructor of the Servo object sets the index of this array
    srvs[jack-1].attach(port[jack+3].pin, SERVO_PULSE_MIN, SERVO_PULSE_MAX);
}

void FSmotor::endServo(byte jack) {
    srvs[jack-1].detach();
}

void FSmotor::setServoMinMax(byte jack, int pulseMin, int pulseMax) {
    srvs[jack-1].attach(port[jack+3].pin, pulseMin, pulseMax);
}

void FSmotor::setServoAngle(byte jack, int angle) {
    srvs[jack-1].write(angle);
}

void FSmotor::setServoPulse(byte jack, int pulse) {
    srvs[jack-1].writeMicroseconds(pulse);
}

void FSmotor::enableDC(byte jack) {
    if (jack==1) {
        digitalWriteFS(ENABLE_DCM1,1);
    } else if (jack==2) {
        digitalWriteFS(ENABLE_DCM2,1);
    }
}

void FSmotor::disableDC(byte jack) {
    if (jack==1) {
        digitalWriteFS(ENABLE_DCM1,0);
    } else if (jack==2) {
        digitalWriteFS(ENABLE_DCM2,0);
    }
}

void FSmotor::setSpeedDC(byte jack, byte value) {
    if (jack==1) {
        analogWriteFS(ENABLE_DCM1, value);
    } else if (jack==2) {
        analogWriteFS(ENABLE_DCM2, value);
    }
}

void FSmotor::dcMotor00(byte jack) {
    if (jack==1) {
        write2bit(DCM1_PLUS,0,DCM1_MINU,0);
    } else if (jack==2) {
        write2bit(DCM2_PLUS,0,DCM2_MINU,0);
    }
}

void FSmotor::dcMotor11(byte jack) {
    if (jack==1) {
        write2bit(DCM1_PLUS,1,DCM1_MINU,1);
    } else if (jack==2) {
        write2bit(DCM2_PLUS,1,DCM2_MINU,1);
    }
}

void FSmotor::dcMotorCW(byte jack) {
    if (jack==1) {
        write2bit(DCM1_PLUS,1,DCM1_MINU,0);
    } else if (jack==2) {
        write2bit(DCM2_PLUS,1,DCM2_MINU,0);
    }
}

void FSmotor::dcMotorCCW(byte jack) {
    if (jack==1) {
        write2bit(DCM1_PLUS,0,DCM1_MINU,1);
    } else if (jack==2) {
        write2bit(DCM2_PLUS,0,DCM2_MINU,1);
    }
}

// DEPRECATED functions

void FSmotor::pinMode(byte pin, byte mode) {
    setPinMode(pin, mode);
}

void FSmotor::digitalWrite(byte pin, byte value) {
    writeDigitalValue(pin, value);
}

byte FSmotor::digitalRead(byte pin) {
    return readDigitalValue(pin);
}
