/*
  FStempHum.cpp - FreeSensors library for Arduino.
  Copyright (C) 2019 Mario A. Guerra. All rights reserved.
  This file is part of FreeSensors, licensed under GNU GPL version 3.
*/

#include "FStempHum.h"

#define WAIT_SENSOR       55             // µs
#define FAIL_TIMEOUT      0xffffffffUL   // arbitrary value

FStempHum::FStempHum() {
    mtype = fmtDIO;
    temp = NAN;
    hum = NAN;
    setType(DHT11);     // DHT11 is the default
    errCount = 0;
}

void FStempHum::begin() {
    toShield->setPinMode(getPortJack(), INPUT_PULLUP);
    maxCiclos = microsecondsToClockCycles(1000); // 1 ms timeout when reading pulses
    prevTime = millis() - minReadInterval;
}

/* Sets the DHT sensor type, and set the minimun read interval, which is 1 sec
 * for DHT11 (max fequency of 1 Hz) and 0.5 sec for DHT22 (max frequency of 2 Hz),
 * adding a plus of 20% to this minimum.
 */
void FStempHum::setType(byte xtipo) {
    tipo = xtipo;
    minReadInterval = (tipo<20)? 1000:500;
    minReadInterval = minReadInterval * 1.2;
}

byte FStempHum::getType() {
    return tipo;
}

/* Needs to be called in loop to update the temperature and humidity values.
 * If called more frequently than minimum it simply returns the previous values,
 * and if it fails to read the sensor raw values also returns the previous values.
 */
void FStempHum::updateValues() {
    unsigned long currTime;

    currTime = millis();
    if ((currTime - prevTime) < minReadInterval) return;
    prevTime = currTime;
    if (readRawValues()) {
        errCount = 0;
        switch (tipo) {
            case DHT11:
                hum = data[0] + data[1] * 0.1;
                temp = data[2];
                if (data[3] & 0x80) {
                    temp = -1 - temp;
                }
                temp += (data[3] & 0x0f) * 0.1;
                break;
            case DHT12:
                hum = data[0] + data[1] * 0.1;
                temp = data[2];
                temp += (data[3] & 0x0f) * 0.1;
                if (data[2] & 0x80) {
                    temp *= -1;
                }
                break;
            case DHT21:
            case DHT22:
                hum = ((word)data[0]) << 8 | data[1];
                hum *= 0.1;
                temp = ((word)(data[2] & 0x7F)) << 8 | data[3];
                temp *= 0.1;
                if (data[2] & 0x80) {
                    temp *= -1;
                }
                break;
        }
    } else {
        // After 3 erroneous readings returns NAN
        errCount++;
        if (errCount>3) {
            temp = NAN;
            hum = NAN;
            errCount = 0;
        } else {
            // temp and hum holds their previous values
        }
    }
}

/* Implements the reading protocol for DHT sensors, first sending
 * a LOW pulse and then expecting a START signal from the sensor.
 * If successfull returns 4 bytes of data in then data[] array,
 * where the last byte data[5] is a checksum of the other four.
 */
bool FStempHum::readRawValues() {
    int i, pin;
    byte bit, port;
    uint32_t lowCiclos, highCiclos, ciclos[80];

    for (i=0; i<5; i++) data[i] = 0;
    pin = getPortPin();
    pinMode(pin, INPUT_PULLUP);
    delay(1);
    pinMode(pin, OUTPUT);
    digitalWrite(pin, LOW);
    switch(tipo) {
        case DHT11:
        case DHT12:
            delay(20); // at least 18 ms
            break;
        case DHT21:
        case DHT22:
            delayMicroseconds(1100); // at least 1 ms
            break;
    }
    pinMode(pin, INPUT_PULLUP);
    delayMicroseconds(WAIT_SENSOR); // let sensor pull line low
    bit = digitalPinToBitMask(pin);
    port = digitalPinToPort(pin);
    // ---- begin time-critical section ----
    noInterrupts();
    // Expect a START signal composed by a LOW for ~80 µs followed by a HIGH for ~80 µs
    if (readPulse(bit,port,LOW) == FAIL_TIMEOUT) {
        interrupts();
        return false;
    }
    if (readPulse(bit,port,HIGH) == FAIL_TIMEOUT) {
        interrupts();
        return false;
    }
    // Now read 40 bits sent by the DHT sensor, each bit is sent as
    // a 50 µs low pulse followed by a variable length high pulse
    for (i=0; i<80; i+=2) {
        ciclos[i] = readPulse(bit,port,LOW);
        ciclos[i+1] = readPulse(bit,port,HIGH);
    }
    interrupts();
    // ---- end time-critical section ----
    for (i=0; i<40; ++i) {
        lowCiclos = ciclos[2*i];
        highCiclos = ciclos[2*i + 1];
        if (lowCiclos == FAIL_TIMEOUT) return false;
        if (highCiclos == FAIL_TIMEOUT) return false;
        data[i / 8] <<= 1;
        if (highCiclos > lowCiclos) {
            data[i / 8] |= 1;
        } else {
            // high cycles are less than the 50 µs low cycle count,
            // so this must be a zero. No need to change data.
        }
    }
    DBG_PRINT("raw data[0] = "); DBG_PRINTLN(data[0], HEX);
    DBG_PRINT("raw data[1] = "); DBG_PRINTLN(data[1], HEX);
    DBG_PRINT("raw data[2] = "); DBG_PRINTLN(data[2], HEX);
    DBG_PRINT("raw data[3] = "); DBG_PRINTLN(data[3], HEX);
    DBG_PRINT("raw data[4] = "); DBG_PRINTLN(data[4], HEX);
    if (data[4] == ((data[0] + data[1] + data[2] + data[3]) & 0xFF)) {
        return true;  // checksum matches
    } else {
        DBG_PRINTLN("checksum failure !");
        return false;
    }
}

unsigned long FStempHum::readPulse(byte bit, byte port, bool level) {
    unsigned long count;
    byte portState;

    count = 0;
    portState = level ? bit : 0;
    while ((*portInputRegister(port) & bit) == portState) {
        if (count++ >= maxCiclos) {
            return FAIL_TIMEOUT; // fails when exceeding the 1 ms timeout
        }
    }
    return count;
}
