/*
  FSultrasonic.cpp - FreeSensors library for Arduino
  Copyright (C) 2018 Mario A. Guerra. All rights reserved.
  License: GNU GPL v3 http://www.gnu.org/licenses/gpl.html

  UltrasonicFS for FreeSensors (with "one pin" operation)
  is based on NewPing Library - vers 1.8 - 07/30/2016
  Created by Tim Eckel - teckel@leethost.com
  Copyright 2016 License: GNU GPL v3 http://www.gnu.org/licenses/gpl.html
  Project home: https://bitbucket.org/teckel12/arduino-new-ping/wiki/Home
  Blog: http://arduino.cc/forum/index.php/topic,106043.0.html
*/

#include "FSultrasonic.h"

FSultrasonic::FSultrasonic() {
    mtype = fmtDIO;
}

void FSultrasonic::begin() {
    initialize(MAX_ULTRA_DISTANCE);
}

void FSultrasonic::initialize(unsigned int maxDistanceCM) {
    uint8_t trigger_pin = getPortPin();
    uint8_t echo_pin = trigger_pin;
    _triggerBit = digitalPinToBitMask(trigger_pin); // Get the port register bitmask for the trigger pin
    _echoBit = digitalPinToBitMask(echo_pin);       // Get the port register bitmask for the echo pin

    _triggerOutput = portOutputRegister(digitalPinToPort(trigger_pin)); // Get the output port register for the trigger pin
    _echoInput = portInputRegister(digitalPinToPort(echo_pin));         // Get the input port register for the echo pin

    _triggerMode = (uint8_t *) portModeRegister(digitalPinToPort(trigger_pin)); // Get the port mode register for the trigger pin
    setMaxDistance(maxDistanceCM); // Call function to set the max sensor distance
}

boolean FSultrasonic::ping_trigger() {
    *_triggerMode |= _triggerBit;      // Set trigger pin to output
    *_triggerOutput &= ~_triggerBit;   // Set the trigger pin low, should already be low, but this will make sure it is
    delayMicroseconds(4);              // Wait for pin to go low
    *_triggerOutput |= _triggerBit;    // Set trigger pin high, this tells the sensor to send out a ping
    delayMicroseconds(10);             // Wait long enough for the sensor to realize the trigger pin is high. Sensor specs say to wait 10 µS
    *_triggerOutput &= ~_triggerBit;   // Set trigger pin back to low
    *_triggerMode &= ~_triggerBit;     // Set trigger pin to input (on one Arduino pin, this is setting the echo pin to input as both are tied to the same Arduino pin)
    if (*_echoInput & _echoBit) return false;               // Previous ping hasn't finished, abort
    _max_time = micros() + _maxEchoTime + MAX_SENSOR_DELAY; // Maximum time we'll wait for ping to start (most sensors are < 450 uS, the SRF06 can take up to 34,300 µS!)
    while (!(*_echoInput & _echoBit)) {                     // Wait for ping to start
        if (micros() > _max_time) return false;             // Took too long to start, abort
    }
    _max_time = micros() + _maxEchoTime; // Ping started, set the time-out
    return true;                         // Ping started successfully
}

unsigned int FSultrasonic::ping() {
    if (!ping_trigger()) return NO_ECHO;          // Trigger a ping, if it returns false, return NO_ECHO to the calling function
    while (*_echoInput & _echoBit) {              // Wait for the ping echo
        if (micros() > _max_time) return NO_ECHO; // Stop the loop and return NO_ECHO (false) if we're beyond the set maximum distance
    }
    return (micros() - (_max_time - _maxEchoTime) - PING_OVERHEAD); // Calculate ping time, include overhead
}

unsigned long FSultrasonic::pingCM() {
    unsigned long echoTime = ping();     // Calls the ping method and returns with the ping echo distance in µS
    return (echoTime / US_ROUNDTRIP_CM); // Call the ping method and returns the distance in centimeters (no rounding)
}

unsigned long FSultrasonic::pingMedian(uint8_t it) {
    unsigned int uS[it], last;
    uint8_t j, i = 0;
    unsigned long t;
    uS[0] = NO_ECHO;

    while (i < it) {
        t = micros();    // Start ping timestamp
        last = ping();   // Send ping

        if (last != NO_ECHO) {          // Ping in range, include as part of median
            if (i > 0) {                // Don't start sort till second ping
                for (j = i; j > 0 && uS[j - 1] < last; j--) { // Insertion sort loop
                    uS[j] = uS[j - 1];  // Shift ping array to correct position for sort insertion
                }
            } else {
                j = 0;                  // First ping is sort starting point
            }
            uS[j] = last;               // Add last ping to array in sorted position
            i++;                        // Move to next ping.
        } else {
            it--;                       // Ping out of range, skip and don't include as part of median
        }
        if (i < it && micros() - t < PING_MEDIAN_DELAY) {
            delay((PING_MEDIAN_DELAY + t - micros()) / 1000); // Millisecond delay between pings
        }
    }
    return (uS[it >> 1]); // Return the ping distance median
}

void FSultrasonic::setMaxDistance(unsigned int maxDistanceCM) {
    _maxEchoTime = min(maxDistanceCM + 1, (unsigned int) MAX_ULTRA_DISTANCE + 1) * US_ROUNDTRIP_CM; // Calculate the maximum distance in µS (no rounding)
}

int FSultrasonic::getMaxDistance() {
    return (_maxEchoTime / US_ROUNDTRIP_CM);  // Call the ping method and returns the distance in centimeters (no rounding)
}

int FSultrasonic::getDistance() {
    int dist;
    dist = pingCM();
    // when there is no echo until max distance, return this max distance plus 1 cm
    if (dist==0) dist = getMaxDistance() + 1;
    return dist;
}
